/* Make docstring objects code font */
span.doc.doc-object-name {
  font-weight: 400;
  font-family: var(--md-code-font-family);
}

h2.doc.doc-heading {
  border-bottom-style: solid;
  border-color: var(--md-typeset-table-color);
  border-width: .05rem;
}

h3.doc.doc-heading {
  border-bottom-style: solid;
  border-color: var(--md-typeset-table-color);
  border-width: .05rem;
}

/* Object level indentation and vertical border */
div.doc-contents:not(.first) {
  padding-left: 25px;
  border-left: .05rem solid var(--md-typeset-table-color);
}

/* Fancier color for operators such as * and |. */
.doc-signature .o {
  color: var(--md-code-hl-special-color);
}

/* Fancier color for constants such as None, True, and False. */
.doc-signature .kc {
  color: var(--md-code-hl-constant-color);
}

a.external:hover::after,
a.autorefs-external:hover::after {
  background-color: var(--md-accent-fg-color);
}
